# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: testTrustedFSObjects.py,v 1.2 2019/08/16 07:02:44 dieter Exp $
from .TestBase import TestCase, getSuite

class TestTrustedFSObjects(TestCase):
  def testPageTemplate(self):
    fi = self.folder.Files
    fi.pt() # we succeed unless we get an exception

  def testPythonScript(self):
    fi = self.folder.Files
    self.assertEqual(fi.py().rstrip(), 'test')



# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: TrustedExpression.py,v 1.2 2019/08/16 07:02:44 dieter Exp $

from sys import modules

from dm.reuse import rebindFunction

import OFS
from Products.PageTemplates.Expressions import \
     ZopePathExpr, PathIterator, \
     getEngine, createZopeEngine, \
     render
from zope.tales.expressions import SimpleModuleImporter
from zope.traversing.adapters import traversePathElement

from . import PythonExpr


ModuleImporter = SimpleModuleImporter()


# sad that we must duplicate this code here.
def boboAwareZopeTraverse(object, path_items, econtext):
    """Traverses a sequence of names, first trying attributes then items.

    This uses Zope 3 path traversal where possible and interacts
    correctly with objects providing OFS.interface.ITraversable when
    necessary (bobo-awareness).
    """
    request = getattr(econtext, 'request', None)
    path_items = list(path_items)
    path_items.reverse()

    while path_items:
        name = path_items.pop()
        if OFS.interfaces.ITraversable.providedBy(object):
          if name.startswith('_'):
            # We know that even 'unrestrictedTraverse' will reject it.
            # We do not want to reimplement 'unrestrictedTraverse' (without
            # the stupid restriction) but still support at least
            # attribute access (maybe, we do more later)
            object = getattr(object, name)
          else:
            object = object.unrestrictedTraverse(name)
        else:
          object = traversePathElement(object, name, path_items,
                                         request=request)
    return object

render = rebindFunction(render,
                        ZRPythonExpr = PythonExpr,
                        )


class ZopePathExpr(ZopePathExpr):
  _TRAVERSER = staticmethod(boboAwareZopeTraverse)
  __init__ = rebindFunction(ZopePathExpr.__init__,
                            boboAwareZopeTraverse = boboAwareZopeTraverse,
                            )
  _eval = rebindFunction(ZopePathExpr._eval,
                         render = render,
                         )

class PathIterator(PathIterator):
  same_part = rebindFunction(PathIterator.same_part,
                             boboAwareZopeTraverse = boboAwareZopeTraverse,
                             )

createZopeEngine = rebindFunction(createZopeEngine,
                                  ZopePathExpr = ZopePathExpr,
                                  PathIterator = PathIterator,
                                  ZRPythonExpr = PythonExpr,
                                  SecureModuleImporter = ModuleImporter,
                                  )

try: _engine = createZopeEngine(ZopePathExpr)
except TypeError: # old version
    _engine = createZopeEngine()
def getEngine(): return _engine





# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: TrustedPageTemplateFile.py,v 1.2 2019/08/16 07:02:44 dieter Exp $
'''A 'PageTemplateFile' without security restrictions.'''

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from dm.reuse import rebindFunction
from .TrustedExpression import getEngine, ModuleImporter


class TrustedPageTemplateFile(PageTemplateFile):
  pt_getEngine = rebindFunction(PageTemplateFile.pt_getEngine,
                                  getEngine = getEngine
                                  )

  pt_getContext = rebindFunction(PageTemplateFile.pt_getContext,
                                 SecureModuleImporter=ModuleImporter,
                                 )




# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: pymixin_2.py,v 1.1 2019/08/16 07:02:44 dieter Exp $
'''Auxiliary mixin class to implement trusted PythonScripts.'''
import __builtin__
from operator import getitem

from ExtensionClass import Base
from RestrictedPython.RestrictionMutator import \
     RestrictionMutator as RM, FuncInfo, _print_target_name
from RestrictedPython.RCompile import RFunction, compileAndTuplize
from RestrictedPython import PrintCollector

from Products.PythonScripts.PythonScript import PythonScript

from dm.reuse import rebindFunction

_Globals = {
  '__builtins__':__builtin__.__dict__,
  '_apply_':      apply,
  '_getitem_':    getitem,
  '_getiter_':    iter,
   '_print_':     PrintCollector,
  '_write_':      lambda ob:ob,
  }
  

class TrustedPythonScriptMixin(Base):
  _newfun = rebindFunction(PythonScript._newfun.im_func,
                           safe_builtins=__builtin__.__dict__,
                           _getattr_=getattr,
                           _getitem_=getitem,
                           _write_=lambda ob:ob,
                           # for Zope 2.7.1 and up
                           get_safe_globals = _Globals.copy,
                           guarded_getattr = getattr,
                           )

  def _compiler(self, *args, **kw):
    # Bah: apparently someone renamed 'globalize' to 'globals' in 'Rfunction.__init__' near Zope 2.7.1
    try:
      gen = RFunction(*args, **kw)
    except TypeError:
      #work around the incompatible change
      if 'globalize' not in kw: raise
      kw['globals'] = kw['globalize']
      del kw['globalize']
      gen = RFunction(*args, **kw)
    gen.rm = RestrictionMutator()
    return compileAndTuplize(gen)


class RestrictionMutator:
  '''a 'RestrictionMutator' without restrictions.'''

  def __init__(self):
    self.warnings = []
    self.errors = []
    self.used_names = {}
    self.funcinfo = FuncInfo()

  def checkName(self, node, name):
    if name == "printed":
      self.error(node, '"printed" is a reserved name.')
    
  
  visitName = RM.visitName.im_func
  visitAssName = RM.visitAssName.im_func
  visitFunction = RM.visitFunction.im_func
  visitModule = RM.visitModule.im_func
  visitClass = RM.visitClass.im_func
  visitImport = RM.visitImport.im_func

  prepBody = RM.prepBody.im_func

  def visitPrint(self, node, walker):
    """we add the current print target if no target is specified."""
    node = walker.defaultVisitNode(node)
    self.funcinfo.print_used = 1
    self.funcinfo._print_used = 1 # compatibility?
    if node.dest is None: node.dest = _print_target_name
    return node

  visitPrintnl = visitPrint 



# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: TrustedFSPageTemplate.py,v 1.3 2019/08/16 07:02:44 dieter Exp $
'''FSPageTemplate unrestricted by Zopes security'''

import sys

from Products.CMFCore.FSPageTemplate import \
     FSPageTemplate, \
     PageTemplate, \
     registerFileExtension

from dm.reuse import rebindFunction
from .Utils import _UnCustomizable
from .TrustedExpression import getEngine, ModuleImporter

class TrustedFSPageTemplate(_UnCustomizable, FSPageTemplate):
  
  meta_type = 'Trusted Filesystem Page Template'

  pt_getEngine = staticmethod(getEngine)

  pt_getContext = rebindFunction(FSPageTemplate.pt_getContext,
                                 SecureModuleImporter=ModuleImporter,
                                 )


registerFileExtension('xpt', TrustedFSPageTemplate)



# Copyright (C) 2004 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: TestBase.py,v 1.2 2019/08/16 07:02:44 dieter Exp $
'''Test base class.

The 'TrustedExecutables' tests use 'ZopeTestCase'.
You must install this package separately (under 'lib/python/Testing').
You can get 'ZopeTestCase' from Zope.org.
'''

from os.path import join
from sys import modules
from unittest import TestSuite, makeSuite

from App.version_txt import getZopeVersion
from Testing.ZopeTestCase import ZopeTestCase, installProduct
from Testing.ZopeTestCase.layer import ZopeLite
from zope import component
from zope.component import globalregistry, _api
from zope.component.globalregistry import \
     globalSiteManager, BaseGlobalComponents
from zope.component.hooks import getSite, setSite
from zope.configuration.xmlconfig import file as load_config

from Products.CMFCore.DirectoryView import registerDirectory
from Products.CMFCore.utils import getContainingPackage

prefix = getContainingPackage(__name__)

installProduct('CMFCore', 1)
installProduct('Five', 1)
registerDirectory('Files', globals())

class Layer(ZopeLite):
  context = None

  @classmethod
  def setUp(cls):
    cls.reg = globalSiteManager
    # set up new component registry
    globalregistry.base = globalregistry.globalSiteManager = BaseGlobalComponents("base")
    # `_api` caches `base` -- invalidate
    _api.base = None
    # in a Plone environment `setSite` has been called and cached `base`
    #   reset
    cls.site = getSite()
    setSite(None)
    if getZopeVersion()[0] == 2:
      # initialize
      cls.context = load_config("meta.zcml", component, cls.context)
    cls.context = load_config("configure.zcml", modules[prefix], cls.context)

  @classmethod
  def tearDown(cls):
    # restore former registry
    globalregistry.base = globalregistry.globalSiteManager = cls.reg
    # `_api` caches `base` -- invalidate
    _api.base = None
    setSite(cls.site)
    cls.context = None
    

class TestCase(ZopeTestCase):
  layer = Layer

  def afterSetUp(self):
    folder = self.folder
    folder.manage_addProduct['CMFCore'].manage_addDirectoryView(
      prefix + ':Files', 'Files'
      )


def getSuite(*testClasses, **kw):
  prefix= kw.get('prefix','test')
  return TestSuite([makeSuite(cl,prefix) for cl in testClasses])
  



# Copyright (C) 2004 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: __init__.py,v 1.1.1.1 2008/05/18 14:00:04 dieter Exp $
'''ZopeTestCase based test suite.'''



# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: TrustedPythonScript.py,v 1.2 2019/08/16 07:02:44 dieter Exp $
'''Python Script unrestricted by Zopes security.

CAUTION: Almost surely, you do not want to make this available.
'''

from Products.PythonScripts.PythonScript import PythonScript

from .TrustedPythonScriptMixin import TrustedPythonScriptMixin

class TrustedPythonScript(TrustedPythonScriptMixin, PythonScript):
  '''Python Script unrestriced by Zopes security.'''
  meta_type = 'Trusted Python Script'



# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: Utils.py,v 1.2 2019/08/16 07:02:44 dieter Exp $
'''Utilities'''

from ExtensionClass import Base

class _UnCustomizable(Base):
  '''mixin class to prevent customization.'''
  def manage_doCustomize(self):
    "do not allow customization"
    raise TypeError('This object does not support customization')



# Copyright (C) 2008-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: PythonExpr.py,v 1.3 2019/08/16 07:02:44 dieter Exp $
'''Trusted equivalent of 'ZRPythonExpr'.'''

from dm.reuse import rebindFunction

from zope.tales.pythonexpr import PythonExpr
from DocumentTemplate.DT_Util import TemplateDict

from Products.PageTemplates.ZRPythonExpr import call_with_ns

class Utd(TemplateDict):
  # required by "InstanceDict"
  guarded_getattr = getattr

call_with_ns = rebindFunction(call_with_ns,
                              Rtd = Utd,
                              )






# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: TrustedFSPythonScript.py,v 1.3 2019/08/16 07:02:44 dieter Exp $
'''FSPythonScript unrestricted by Zopes security.'''
import linecache, os

from Products.CMFCore.FSPythonScript import \
     FSPythonScript, \
     registerFileExtension

from dm.reuse import rebindFunction
from .Utils import _UnCustomizable
from .TrustedPythonScriptMixin import TrustedPythonScriptMixin
from .TrustedPythonScript import TrustedPythonScript


# ATT: in case Chris Wither's "zdb" is installed, it is essential
#  that we are imported first. Otherwise,
#  the original "FSPythonScript._write" is lost and can no longer
#  be rebound
_rebound_write = rebindFunction(FSPythonScript._write,
                          PythonScript=TrustedPythonScript,
                          _NCPythonScript=TrustedPythonScript,
                          )


class TrustedFSPythonScript(_UnCustomizable, TrustedPythonScriptMixin, FSPythonScript):
  meta_type = 'Trusted Filesystem Python Script'

  def _write(self, *args, **kw):
    _rebound_write(self, *args, **kw)
    # make it debuggable -- stolen from Chris Wither's "zdb"
    size = len(self._body)
    if not size: return
    filename = self._filepath
    try: stat = os.stat(filename)
    except os.error: return 
    if filename.startswith('pypackage:'):
      try: from pypackage import getLocation
      except ImportError: from packageresources import getLocation
      filename = getLocation(filename)
    lines = [l+'\n' for l in self._body.split('\n')]
    linecache.cache[filename] = stat.st_size, stat.st_mtime, lines, filename




registerFileExtension('xpy', TrustedFSPythonScript)



# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: testTrustedPageTemplateFile.py,v 1.3 2019/08/16 07:02:44 dieter Exp $

from .TestBase import TestCase

from Products.TrustedExecutables.TrustedPageTemplateFile import \
     TrustedPageTemplateFile

class TestTrustedPageTemplateFile(TestCase):
  def testNoUnauthorized(self):
    f = self.folder
    fi = TrustedPageTemplateFile('Files/pt.xpt', globals()).__of__(f)
    fi() # we succeed unless we get an exception

  def testCallPy(self):
    # test that calling a python script from a trusted page template succeeds
    # (this broke in version 1.0.0)
    fi = self.folder.Files
    self.assertEqual(fi.pt_call_py().rstrip(), 'test')



from os.path import abspath, dirname, join
try:
  # try to use setuptools
  from setuptools import setup
  setupArgs = dict(
      install_requires=[
      'dm.reuse',
      "setuptools", # to make `buildout` happy
      # depends on `Zope2/Zope [4]`
      # the `*FS*` executables depend on `Products.CMFCore`
      ],
      include_package_data=True,
      namespace_packages=['Products'],
      zip_safe=False, # to let the tests work
      )
except ImportError:
  # use distutils
  from distutils import setup
  setupArgs = dict(
    )

cd = abspath(dirname(__file__))
pd = join(cd, 'Products', 'TrustedExecutables')

def pread(filename, base=pd): return open(join(base, filename)).read().rstrip()

setup(name='Products.TrustedExecutables',
      version=pread('VERSION.txt').split('\n')[0],
      description='Trusted Zope 2 executables (scripts, templates) -- unrestricted by Zope 2 security. For Zope 2.13 and above',
      long_description=pread('README.txt'),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Framework :: Zope2',
        'Framework :: Zope',
        'Framework :: Zope :: 4',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
        ],
      author='Dieter Maurer',
      author_email='dieter@handshake.de',
      url='http://www.dieter.handshake.de/pyprojects/zope',
      packages=['Products', 'Products.TrustedExecutables', 'Products.TrustedExecutables.tests'],
      keywords='Zope 2, trusted, executables, unrestricted, security ',
      license='BSD (see "Products/TrustedExecutables/LICENSE.txt", for details)',
      **setupArgs
      )






# Copyright (C) 2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: pymixin_4.py,v 1.1 2019/08/16 07:02:44 dieter Exp $
'''Auxiliary mixin class to implement trusted PythonScripts.'''
try: import __builtin__ # Python 2
except ImportError: import builtins as __builtin__ # Python 3
from operator import getitem

from AccessControl.ZopeGuards import _metaclass, protected_inplacevar
from ExtensionClass import Base
from RestrictedPython import PrintCollector
from RestrictedPython.Guards import \
     guarded_iter_unpack_sequence, guarded_unpack_sequence
from RestrictedPython.compile import \
     compile_restricted_function, RestrictingNodeTransformer

from Products.PythonScripts.PythonScript import PythonScript

from dm.reuse import rebindFunction


_Globals = dict(
  __builtins__=__builtin__.__dict__,
  __metaclass__=_metaclass,
  _apply_=lambda f, *args, **kw: f(*args, **kw),
  _getitem_=getitem,
  _getiter_=iter,
  _iter_unpack_sequence_=guarded_iter_unpack_sequence,
  _unpack_sequence_=guarded_unpack_sequence,
   _print_=PrintCollector,
  _write_=lambda ob:ob,
  _inplace_var_=rebindFunction(protected_inplacevar,
                               isinstance=lambda *args: True),
  )
  
class _UnrestrictingNodeTransformer(RestrictingNodeTransformer):
  def error(*args, **kw): pass # everything is allowed

  def generic_visit(self, node):
    return self.node_contents_visit(node) # transform children


class TrustedPythonScriptMixin(Base):
  _newfun = rebindFunction(PythonScript._newfun,
                           get_safe_globals = _Globals.copy,
                           guarded_getattr = getattr,
                           )
  _compile = rebindFunction(
    PythonScript._compile,
    compile_restricted_function=rebindFunction(
      compile_restricted_function,
      argRebindDir=dict(policy=_UnrestrictingNodeTransformer)
      )
                            )



# Copyright (C) 2004-2019 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
# see "LICENSE.txt" for details
#       $Id: __init__.py,v 1.2 2019/08/16 07:02:44 dieter Exp $
'''TrustedExecutables

this product defines a set of executable objects not restricted
by Zopes security checks.

The security checks are expensive. Thus avoiding them can
drastically speed things up.

You should use these features only for file system based objects
which are considered safe according to the Zope security philosophie.

CAUTION: These objects must make their own security checks when
  they access untrusted objects.
'''

from . import TrustedPageTemplateFile

try: import Products.CMFCore
except ImportError: pass
else:
  from . import TrustedFSPageTemplate, TrustedFSPythonScript

# add utilities to "__builtin__"
try: import builtins as __builtin__ # Python 3
except ImportError: import __builtin__ # Python 2
_bd = __builtin__.__dict__
from DocumentTemplate.DT_Util import TemplateDict
for _k,_v in TemplateDict.__dict__.items():
  if not _k.startswith('_') and _k not in _bd: _bd[_k] = _v
del _k; del _v; del _bd



# Copyright (C) 2019 by Dr. Dieter Maurer, Illtalstr. 25, D-66571 Bubach, Germany
# see "LICENSE.txt" for details
#       $Id: TrustedPythonScriptMixin.py,v 1.2 2019/08/16 07:02:44 dieter Exp $
'''Auxiliary mixin class to implement trusted PythonScripts.'''
from App.version_txt import getZopeVersion

if getZopeVersion()[0] >= 4: from .pymixin_4 import TrustedPythonScriptMixin
else: from .pymixin_2  import TrustedPythonScriptMixin
